/*
  Custom pforth 'main' routine that interacts with the Linux kernel
  through the /proc filesystem, by David N. Welton
  <davidw@dedasys.com>.

  See README-kpforth.txt

  */

/* @(#) pf_main.c 98/01/26 1.2 */
/***************************************************************
 ** Forth based on 'C'
 **
 ** main() routine that demonstrates how to call PForth as
 ** a module from 'C' based application.
 ** Customize this as needed for your application.
 **
 ** Author: Phil Burk
 ** Copyright 1994 3DO, Phil Burk, Larry Polansky, Devid Rosenboom
 **
 ** The pForth software code is dedicated to the public domain,
 ** and any third party may reproduce, distribute and modify
 ** the pForth software code or any derivative works thereof
 ** without any compensation or license.  The pForth software
 ** code is provided on an "as is" basis without any warranty
 ** of any kind, including, without limitation, the implied
 ** warranties of merchantability and fitness for a particular
 ** purpose and their equivalents under the laws of any jurisdiction.
**
***************************************************************/

#ifdef PF_NO_STDIO
#ifdef __KERNEL__
#define ERR(msg) { printk(KERN_NOTICE msg) ; }
#define NULL  ((void *) 0)
#else
#define ERR(msg) /* { printf msg; } */
#define NULL  ((void *) 0)
#endif /* __kernel__  */
#else
#include <stdio.h>
#define ERR(msg) { printf msg; }
#endif

#include "pforth.h"
#define DEFAULT_RETURN_DEPTH (512)
#define DEFAULT_USER_DEPTH (512)
#define DEFAULT_HEADER_SIZE (120000)
#define DEFAULT_CODE_SIZE (300000)

#ifdef __MWERKS__
#include <console.h>
#include <sioux.h>
#endif

#ifndef TRUE
#define TRUE (1)
#define FALSE (0)
#endif

#ifndef __KERNEL__
int main( int argc, char **argv )
{
    const char *DicName = "/usr/lib/pforth/pforth.dic";
    const char *SourceName = NULL;
    char IfInit = FALSE;
    char *s;
    int32 i;
    int Result;

/* For Metroworks on Mac */
#ifdef __MWERKS__
    argc = ccommand(&argv);
#endif

/* Parse command line. */
    for( i=1; i<argc; i++ )
    {
	s = argv[i];

	if( *s == '-' )
	{
	    char c;
	    s++; /* past '-' */
	    c = *s++;
	    switch(c)
	    {
	    case 'i':
		IfInit = TRUE;
		DicName = NULL;
		break;
	    case 'q':
		pfSetQuiet( TRUE );
		break;
	    case 'd':
		DicName = s;
		break;
	    default:
		ERR(("Unrecognized option!\n"));
		ERR(("pforth {-i} {-q} {-dfilename.dic} {sourcefilename}\n"));
		Result = 1;
		goto on_error;
		break;
	    }
	}
	else
	{
	    SourceName = s;
	}
    }
/* Force Init */
#ifdef PF_INIT_MODE
    IfInit = TRUE;
    DicName = NULL;
#endif

    Result = pfDoForth( DicName, SourceName, IfInit);

on_error:
    return Result;
}
#else /* __KERNEL__  */

/* LINUX KERNEL STUFF  */

/* Standard in kernel modules */
#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/module.h>   /* Specifically, a module */

/* Deal with CONFIG_MODVERSIONS */
#if 0
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#endif

/* Necessary because we use proc fs */
#include <linux/proc_fs.h>

#include <asm/uaccess.h>

#include "kernel_io.h"

/* Cruft from pforth  */
extern cfTaskData *gCurrentTask;
#define TOUCH(argument) ((void)argument)

/* memory functions  */

char *pfAllocMem( int32 NumBytes )
{
    return kmalloc(NumBytes, GFP_KERNEL);
}
void pfFreeMem( void *Mem )
{
    kfree(Mem);
}

#define SCRIPT_LEN 1024

char *OutScript;
int OutScriptindex = 0;
int OutScriptalloc = SCRIPT_LEN;

char *InScript;
int InScriptindex = 0;

int mycount = 0;

/* The module's file functions ********************** */

int sdTerminalIn()
{
    int c;
    char *addr = InScript + InScriptindex;
    if (addr == NULL)
	return EOF;

    if (*addr == '\0')
    {
//	InScript = NULL;
	return EOF;
    }
    c = *addr;
    InScriptindex ++;
    return c;
}

int sdTerminalOut(char c)
{
    char *newscript;
    OutScriptindex ++;
    if (OutScriptindex >= OutScriptalloc)
    {
	newscript = kmalloc(OutScriptindex + SCRIPT_LEN, GFP_KERNEL);
	newscript = memcpy(newscript, OutScript, OutScriptindex-1);
	kfree (OutScript);
	OutScript = newscript;
	OutScriptalloc += SCRIPT_LEN;
    }
    OutScript[OutScriptindex - 1] = c;
    OutScript[OutScriptindex] = '\0';
    return c;
}

int sdTerminalFlush(void)
{
    return 0;
}

/* Since we use the file operations struct, we can't
 * use the special proc output provisions - we have to
 * use a standard read function, which is this function */
static ssize_t module1_output(
    struct file *file,   /* The file read */
    char *buf, /* The buffer to put data to (in the
                * user segment) */
    size_t len,  /* The length of the buffer */
    loff_t *offset) /* Offset in the file - ignore */
{
    static int finished = 0;
    int i;
    char *message;
    message = kmalloc(InScriptindex + 2, GFP_KERNEL);
    /* We return 0 to indicate end of file, that we have
     * no more information. Otherwise, processes will
     * continue to read from us in an endless loop. */
    if (finished) {
	finished = 0;
	return 0;
    }

    sprintf(message, "%s", InScript);
    for(i=0; i < len && message[i]; i++)
	put_user(message[i], buf+i);

    finished = 1;
    kfree(message);
    return i;  /* Return the number of bytes "read" */
}

static ssize_t module2_output(
    struct file *file,   /* The file read */
    char *buf, /* The buffer to put data to (in the
                * user segment) */
    size_t len,  /* The length of the buffer */
    loff_t *offset) /* Offset in the file - ignore */
{
    static int finished = 0;
    int i;
    char *message;
    message = kmalloc(OutScriptindex + 2, GFP_KERNEL);

    if (finished) {
	finished = 0;
	return 0;
    }

    printk(KERN_INFO "len = %d, offset = %d\n", offset);
    sprintf(message, "%s", OutScript);
    for(i=0; i < len && message[i]; i++)
	put_user(message[i], buf+i);

    finished = 1;
    kfree(message);
    return i;  /* Return the number of bytes "read" */
}

/* This function receives input from the user when the
 * user writes to the /proc file. */
static ssize_t module_input(
    struct file *file,
    const char *buf,
    size_t length,
    loff_t *offset)
{
    int i;

    for(i = 0; i < SCRIPT_LEN - 1 && i < length; i++)
	get_user(InScript[i], buf+i);

    InScript[i] = '\0';
    return i; /* number of bytes written to us  */
}

int module_open(struct inode *inode, struct file *file)
{
    MOD_INC_USE_COUNT;
    file->private_data = kmalloc(sizeof(int), GFP_KERNEL);
    file->private_data = mycount;
    mycount ++;
    return 0;
}

/* I trigger the execution of the script when the file is closed.  I'm
   not sure better way to do this.  */

int module_close(struct inode *inode, struct file *file)
{
    cfTaskData *cftd;
    cfDictionary *dic;
    int32 Result = 0;
    int32 IfInit = 0;
    ExecToken  EntryPoint = 0;
    const char *DicName = "/usr/lib/pforth/pforth.dic";

    /* reset everything */
    kfree(OutScript);
    OutScript = kmalloc(SCRIPT_LEN, GFP_KERNEL);
    *OutScript = '\0';
    OutScriptindex = 0;
    InScriptindex = 0;
    OutScriptalloc = SCRIPT_LEN;

    printk(KERN_WARNING ">>> %d\n", (int)(file->private_data));
    pfInitGlobals();
    /* Allocate forth Task structure. */
    cftd = pfCreateTask( DEFAULT_USER_DEPTH, DEFAULT_RETURN_DEPTH );

    if (cftd)
    {
	pfSetCurrentTask( cftd );
	TOUCH(IfInit);
	{
	    dic = pfLoadDictionary( DicName, &EntryPoint );
	}
	if( dic == NULL ) goto error;

	pfExecByName("AUTO.INIT");

	if(EntryPoint != 0)
	{
	    pfExecuteToken( EntryPoint );
	}
	Result = pfRunForth();
	printk(KERN_INFO "Result = %d\n", Result);
    }

    MOD_DEC_USE_COUNT;

    return 0;  /* success */

error:
    MOD_DEC_USE_COUNT;
    return 1;

}


/* File operations for our proc file. This is where we
 * place pointers to all the functions called when
 * somebody tries to do something to our file. NULL
 * means we don't want to deal with something. */
static struct file_operations KPForth_File_Ops =
{
    read: module1_output, /* "read" from the file */
    write: module_input, /* write to the file  */
    open: module_open,
    release: module_close
};

static struct file_operations KPForth_File_Out_Ops =
{
    read: module2_output, /* "read" from the file */
//    write: module2_input, /* write to the file  */
//    open: module_open,
//    release: module_close
};

/* Module initialization and cleanup ******************* */

/* Initialize the module - register the proc file */
int init_module()
{
    struct proc_dir_entry *KPForth_Proc_File = NULL;
    struct proc_dir_entry *KPForth_Proc_Out_File = NULL;
    InScript = kmalloc(SCRIPT_LEN, GFP_KERNEL);
    OutScript = kmalloc(SCRIPT_LEN, GFP_KERNEL);
    *OutScript = '\0';

    KPForth_Proc_File = create_proc_entry("kpforth", S_IFREG | S_IRUGO | S_IWUSR, &proc_root);
    KPForth_Proc_Out_File = create_proc_entry("kpforth-out", S_IFREG | S_IRUGO | S_IWUSR, &proc_root);

    KPForth_Proc_File->data = (void *)KPForth_Proc_File;
    KPForth_Proc_File->proc_fops = &KPForth_File_Ops;
    KPForth_Proc_File->size = 80;
    KPForth_Proc_File->nlink = 1;

    KPForth_Proc_Out_File->data = (void *)KPForth_Proc_Out_File;
    KPForth_Proc_Out_File->proc_fops = &KPForth_File_Out_Ops;
    KPForth_Proc_Out_File->size = 80;
    KPForth_Proc_Out_File->nlink = 1;

    printk(KERN_NOTICE "kpforth module loaded\n");
    return 0;
}

/* Cleanup - unregister our file from /proc */
void cleanup_module()
{
    kfree(OutScript);
    kfree(InScript);
    printk(KERN_NOTICE "kpforth module unloaded\n");
    remove_proc_entry("kpforth", &proc_root);
}

#endif
